local function find_zephyr()
	local val = utils:list_reg_value([[HKCU\Software\Kitware\CMake\Packages\Zephyr]]);
	if val and #val >= 1 then
		local path = utils:read_reg([[HKCU\Software\Kitware\CMake\Packages\Zephyr\]]..val[1]);
		return make_path(path, "../../.."), make_path(path);
	end;
	local home_zephyr = make_path(os.getenv("USERPROFILE") or "", "zephyrproject/zephyr");
	if utils:file_time(home_zephyr) then
		return home_zephyr, make_path(home_zephyr, "share/zephyr-package/cmake");
	end;
	home_zephyr = make_path(os.getenv("ZEPHYR_BASE"));
	if utils:file_time(home_zephyr) then
		return home_zephyr, make_path(home_zephyr, "share/zephyr-package/cmake");
	end;
end;

local function find_zephyr_sdk(home)
	local dirs = utils:list_dir(make_path(home or "", "zephyr-sdk-*"), true);
	if dirs then
		return make_path(home, dirs[1]);
	end;
	local zephyr_sdk_dir = make_path(os.getenv("ZEPHYR_SDK_INSTALL_DIR") or "");
	if utils:file_time(zephyr_sdk_dir) then
		return zephyr_sdk_dir;
	end;
	local userprofile = os.getenv("USERPROFILE") or "";
	dirs = utils:list_dir(make_path(userprofile, "zephyr-sdk-*"), true);
	if dirs then
		return make_path(userprofile, dirs[1]);
	end;
	local programfiles = os.getenv("PROGRAMFILES") or "";
	dirs = utils:list_dir(make_path(programfiles, "zephyr-sdk-*"), true);
	if dirs then
		return make_path(programfiles, dirs[1]);
	end;
end;

local zephyr_base_path, zephyr_cmake_path = find_zephyr();
local zephyr_sdk_path = zephyr_base_path and find_zephyr_sdk(make_path(zephyr_base_path, "..")) or "";

local function find_zephyr_toolset()
	local zephyr_base_path, zephyr_cmake_path = find_zephyr();
	if not zephyr_base_path then
		return;
	end;
	local zephyr_home = make_path(zephyr_base_path, "..");
	local zephyr_sdk_path = find_zephyr_sdk(zephyr_home);
	if not zephyr_sdk_path then
		print("Can NOT find Zephyr SDK path!");
		return;
	end;
	local virtual_env = make_path(zephyr_home, ".venv/Scripts");
	local west_bin = make_path(virtual_env, "west.exe");
	if utils:file_time(west_bin) == nil then
		print("Can NOT find west.exe at ", west_bin, "!");
		return;
	end;
	local tool_env = {
		["ZEPHYR_BASE"] = zephyr_base_path;
		["ZEPHYR_SDK_INSTALL_DIR"] = zephyr_sdk_path;
		["VIRTUAL_ENV"] = virtual_env;
	};
	local debug_adapter = toolset_make_debug_adapter({"JLink", "JLink(OpenOCD)", "T-Head DebugServer", "pyOCD"});
	local zephyr_toolset = {
		["type"] = "zephyr";
		["sdk_type"] = "zephyr";
		["name"] = "zephyr";
		-- ["version"] = string.format("%u.%u.%u", esp_ver_major, esp_ver_minor, esp_ver_patch);
		-- ["version_number"] = esp_ver_major*0x100000 + esp_ver_minor*0x1000 + esp_ver_patch;
		["home"] = make_path(zephyr_home);
		["envs"] = tool_env;
		["bin"] = {
			virtual_env;
		};
		["include"] = {};
		["lib"] = {};
		-- ["targets"] = table.keys(zephyr_targets);
		-- ["boards"] = zephyr_boards;
		["tools"] = {
			["debugger"] = "zephyr";
			-- ["openocd"] = zephyr_openocd;
			-- ["cmake"] = zephyr_cmake;
			-- ["make"] = zephyr_make;
			-- ["ninja"] = ninja_path;
			["west"] = west_bin;
			["gcc"] = true;
			["g++"] = true;
			-- ["cxx"] = zephyr_gpp;
			-- ["cc"] = zephyr_gcc;
			-- ["gdb"] = zephyr_gdb;
			-- ["gcc-pretty-print"] = make_path(toolchain_root, "share/gcc-"..zephyr_gcc_ver.."/python");
			["vcpkg"] = "DISABLED";
		};
		-- ["debug_adapter"] = debug_adapter;
		encoding = "utf-8";
	};
	local all_probes = {
		["jlink"] = {
			["runner"] = "jlink";
			["type"] = "zephyr";
			["transports"] = {};
			["speeds"] = {};
		};
		["openocd"] = {
			["runner"] = "openocd";
			["type"] = "zephyr";
			["transports"] = {};
			["speeds"] = {};
		};
		["pyocd"] = {
			["runner"] = "pyocd";
			["type"] = "zephyr";
			["transports"] = {};
			["speeds"] = {};
		};
	};
	local available_probes = {
		["auto"] = {
			["runner"] = "auto";
			["type"] = "zephyr";
			["transports"] = {};
			["speeds"] = {};
		};
	};
	for name, adapter in pairs(debug_adapter.adapters) do
		if adapter["type"] == "openocd" or adapter["type"] == "pyocd" or adapter["type"] == "jlink" then
			table.insert(zephyr_toolset.bin, make_path(adapter.bin, ".."));
			available_probes[adapter["type"]] = all_probes[adapter["type"]];
		end;
	end;
	zephyr_toolset.bin = table.unique(zephyr_toolset.bin);
	zephyr_toolset["debug_adapter"] = toolset_make_debug_adapter(
		{},
		available_probes
	);
	-- table.sort(zephyr_toolset.targets);
	print("Zephyr found at ", zephyr_base_path);
	print("\tSDKs: ", zephyr_sdk_path);
	print("\twest: ", west_bin);
	return zephyr_toolset;
end;

ZEPHYR_TOOLSET = find_zephyr_toolset();
ZEPHYR_CMAKE_PROJECT = false;

function zephyr_create_project(path, name, target, on_finished_handler)
	if not ZEPHYR_TOOLSET then
		return false;
	end;
	local toolset = ZEPHYR_TOOLSET;
	local project_name = name:gsub("%s", "-");
local cmake_list = [[cmake_minimum_required(VERSION 3.15)

# SELECT YOUR BOARD HERE!
set(BOARD rpi_pico)

find_package(Zephyr REQUIRED HINTS $ENV{ZEPHYR_BASE})

project(%s)

set(CMAKE_C_STANDARD 11)
set(CMAKE_CXX_STANDARD 14)

# include_directories(<HEADER_PATH>)

# add_link_options(<LINKER_FLAGS>)

# add_compile_options(<COMPILER_FLAGS>)

file(GLOB APP_FILES
	LIST_DIRECTORIES false
	RELATIVE ${CMAKE_CURRENT_SOURCE_DIR}
	${CMAKE_CURRENT_SOURCE_DIR}/*.c
	${CMAKE_CURRENT_SOURCE_DIR}/*.cpp
	${CMAKE_CURRENT_SOURCE_DIR}/*.cc
	${CMAKE_CURRENT_SOURCE_DIR}/*.cxx
)

target_sources(app PRIVATE ${APP_FILES})

# target_precompile_headers(${TARGET_NAME} PRIVATE pch_header.h)

# target_link_libraries(${TARGET_NAME}
#	<LIB_NAME>
# )

]];

local prj_conf = [[]];

local main_c = [[#include <stdio.h>

int main(void) {
	printf("Hello Zephyr!\n");
	return 0;
}
]];
	write_file(make_path(path, "CMakeLists.txt"), cmake_list:format(project_name));
	write_file(make_path(path, "prj.conf"), prj_conf);
	write_file(make_path(path, "main.c"), main_c);
	if on_finished_handler then
		local extras = "\t[\".extra\"] = {\n"..
		"\t\t[\"com_baud\"] = [[2048000]];\n"..
		"\t\t[\"mon_baud\"] = [[2048000]];\n"..
		"\t}\n";
		on_finished_handler(true, toolset.name, "main.c", extras:format(target));
	end;
end;

-- local old_project_target_filter = project._cmake_project_target_filter;
-- project._cmake_project_target_filter = function(self, target, cmd)
-- 	if ZEPHYR_CMAKE_PROJECT and cmd.toolset and cmd.toolset["type"] == "zephyr" then
-- 		return target["dir"] and (
-- 			target["dir"]:sub(1,18) == "build_out/drivers/" or
-- 			target["dir"]:sub(1,21) == "build_out/components/" or
-- 			target["name"] == "zephyr_sdk_lib" or
-- 			target["name"] == "zephyr_sdk_gcc_lib"
-- 		);
-- 	end;
-- 	return old_project_target_filter(self, target, cmd);
-- end;

local old_project_filter = project._cmake_project_filter;
project._cmake_project_filter = function(self, targets, cmd)
	targets = old_project_filter(self, targets, cmd);
	if ZEPHYR_CMAKE_PROJECT then
		local zephyr_modules_path = make_path(cmd.build_path, "modules");
		local zephyr_targets = {};
		for idx, target in ipairs(targets) do
			if target["type"] ~= 4 and not is_file_in_path(make_path(target.output), zephyr_modules_path) then
				if target["type"] == 2 and (
						target.name == "isr_tables" or
						target.name == "kernel" or
						target.name == "app" or
						target.name == "zephyr" or
						target.name:match("^arch__.*") or
						target.name:match("^drivers__.*") or
						target.name:match("^soc__.*") or
						target.name:match("^lib__.*")
					)
				then
				elseif target.name == "zephyr_pre0" then
				else
					table.insert(zephyr_targets, target);
				end;
			end;
		end;
		-- add install target enable the install menu item
		table.insert(zephyr_targets, {
				["name"] = "install";
				["type"] = 4;
		});
		-- enable flash target
		table.insert(zephyr_targets, {
				["name"] = "flash";
				["type"] = 4;
		});
		return zephyr_targets;
	end;
	return targets;
end;

local old_cmake_find_build_command = project._cmake_find_build_command;
project._cmake_find_build_command = function(self, build_target)
	if not ZEPHYR_CMAKE_PROJECT then
		return old_cmake_find_build_command(self, build_target);
	end;

	local config = menu_bar.get_cmake_config();
	local toolset = ZEPHYR_TOOLSET;
	-- local target = menu_bar.get_cmake_cpu() or "hpm5301";
	local cmake_extra_arguments = nil;

	if build_target == "install" then
		build_target = "flash";
	end;
	local target_board =  "";
	cmake_extra_arguments = "";

	local build_cmd = self:_cmake_find_build_command_impl(build_target, toolset, config, cmake_extra_arguments);
	if build_cmd == nil then
		return nil;
	end;
	build_cmd.environment = table.clone(build_cmd.environment, toolset.envs);
	build_cmd.__zephyr_build_target = build_target;
	return build_cmd;
end;

local old_cmake_gen_do_build_params = project._cmake_gen_do_build_params;
project._cmake_gen_do_build_params = function(self, cmd, build_type, on_complete, prepend_cmds, apppend_cmds)
	if not ZEPHYR_CMAKE_PROJECT then
		return old_cmake_gen_do_build_params(self, cmd, build_type, on_complete, prepend_cmds, apppend_cmds);
	end;
	local toolset = cmd.toolset;
	if cmd.__zephyr_build_target == "flash" then
		local com_port = toolset.envs["BLPORT"];
		local com_port_baud = toolset.envs["BLBAUD"];
		-- if the corresponding port was opened, trying to stop the serial port monitor
		if edx:set_serial_monitor(com_port, false) then
			-- the com port is closed successfully
			local old_on_complete = on_complete;
			on_complete = function(success)
				-- trying to reopen the com port
				edx:set_serial_monitor(com_port, true);
				if old_on_complete then
					old_on_complete(success);
				end;
			end;
		end;
		build_type = "flash";
	end;
	return old_cmake_gen_do_build_params(self, cmd, build_type, on_complete, prepend_cmds, apppend_cmds);
end;

table.insert(xws_mgr._pre_load_cmake_project,
	function(self, dir, cmake_lists, cmake_lists_data)
		ZEPHYR_CMAKE_PROJECT = string.match(cmake_lists_data, [[%s*find_package%(%s*Zephyr%s*]]) ~= nil;
		if not ZEPHYR_CMAKE_PROJECT then
			return false;
		end;
		if ZEPHYR_TOOLSET == nil then
			ZEPHYR_CMAKE_PROJECT = false;
			print("\x1b[33;41m", LANG("Zephyr not found!"),"\x1b[0m");
			return false;
		end;
		project.show_command_line = nil;
		local zephyr_toolset = ZEPHYR_TOOLSET;
		local def_buadrate = project:get_extra_setting("com_baud", SERIAL_DEFAULT_BAUDRATE);
		local mon_buadrate = project:get_extra_setting("mon_baud", SERIAL_DEFAULT_BAUDRATE);
		menu_bar.update_cmake_toolset({ZEPHYR_TOOLSET});
		menu_bar.set_cmake_toolset(zephyr_toolset.name);
		menu_bar.show_cmake_items(
			"toolset",
			"serial_ports",
			"serial_baud"
			-- "cpus"
			-- "boards",
		);
		menu_bar.set_cmake_text(LANG("menu_bar/cmake/zephyr_sdk"));
		-- menu_bar.update_cmake_cpu(zephyr_toolset.targets);
		-- menu_bar.set_cmake_cpu(def_target);
		menu_bar.update_cmake_serial_baud(
			SERIAL_BAUDRATES_HI,
			def_buadrate
		);
		menu_bar.update_cmake_serial_port(
			function()
				local ports = {utils:list_serial_ports()};
				table.sort(ports, function(a, b)
						return tonumber(a[1]:sub(4)) < tonumber(b[1]:sub(4));
				end);
				return ports;
			end,
			nil,
			function(val) return val[1], table.concat(val, "\t"); end
		);
		edx:update_serial_baudrate(SERIAL_BAUDRATES, mon_buadrate);
		return true;
	end
);

table.insert(xws_mgr._pre_unload_dir,
	function(self, dir)
		if not ZEPHYR_CMAKE_PROJECT then
			return false;
		end;
		-- project:set_extra_setting("target_cpu", menu_bar.get_cmake_cpu());
		return false;
	end
);

table.insert(xws_mgr._post_unload_dir,
	function(self, dir)
		if ZEPHYR_CMAKE_PROJECT then
			menu_bar.hide_cmake_items("serial_ports", "serial_baud", "debug_adapters", "debug_adapter_speed", "debug_adapter_transport", "cpus", "boards");
			menu_bar.show_cmake_items("toolset");
			menu_bar.update_cmake_toolset(TOOLSETS);
			menu_bar.set_cmake_text(LANG("menu_bar/cmake"));
			project.show_command_line = nil;
		end;
		ZEPHYR_CMAKE_PROJECT = false;
		return false;
	end
);
